// This file is part of the Polymesh distribution (https://github.com/PolymathNetwork/Polymesh).
// Copyright (c) 2020 Polymath

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3.

// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

//! # Identity module
//!
//! This module is used to manage identity concept.
//!
//!  - [Module](./struct.Module.html)
//!  - [Trait](./trait.Trait.html)
//!
//! ## Overview :
//!
//! Identity concept groups different account (keys) in one place, and it allows each key to
//! make operations based on the constraint that each account (permissions and key types).
//!
//! Any account can create and manage one and only one identity, using
//! [register_did](./struct.Module.html#method.register_did). Other accounts can be added to a
//! target identity as signing key, where we also define the type of account (`External`,
//! `MultiSign`, etc.) and/or its permission.
//!
//! Some operations at identity level are only allowed to its administrator account, like
//! [set_master_key](./struct.Module.html#method.set_master_key) or
//!
//! ## Identity information
//!
//! Identity contains the following data:
//!  - `master_key`. It is the administrator account of the identity.
//!  - `signing_keys`. List of keys and their capabilities (type of key and its permissions) .
//!
//! ## Freeze signing keys
//!
//! It is an *emergency action* to block all signing keys of an identity and it can only be performed
//! by its administrator.
//!
//! see [freeze_signing_keys](./struct.Module.html#method.freeze_signing_keys)
//! see [unfreeze_signing_keys](./struct.Module.html#method.unfreeze_signing_keys)
//!
//! ## Interface
//!
//! ### Dispatchable Functions
//!
//! - `register_did` - Register a new did with a CDD claim for the caller.
//! - `cdd_register_did` - Registers a new did for the target and attaches a CDD claim to it.
//! - `invalidate_cdd_claims` - Invalidates any claim generated by `cdd` from `disable_from` timestamps.
//! - `remove_signing_keys` - Removes specified signing keys of a DID if present.
//! - `set_master_key` - Sets a new master key for a DID.
//! - `accept_master_key` - Accept authorization to become the new master key of an identity.
//! - `change_cdd_requirement_for_mk_rotation` - Sets if CDD authorization is required for updating master key of an identity.
//! - `join_identity_as_key` - Join an identity as a signing key.
//! - `join_identity_as_identity` - Join an identity as a signing identity.
//! - `add_claim` - Adds a new claim record or edits an existing one.
//! - `batch_add_claim` - Adds a new batch of claim records or edits an existing one.
//! - `forwarded_call` - Creates a call on behalf of another DID.
//! - `revoke_claim` - Marks the specified claim as revoked.
//! - `batch_revoke_claim` - Revokes multiple claims in a batch.
//! - `set_permission_to_signer` - Sets permissions for an specific `target_key` key.
//! - `freeze_signing_keys` - Disables all signing keys at `did` identity.
//! - `unfreeze_signing_keys` - Re-enables all signing keys of the caller's identity.
//! - `add_authorization` - Adds an authorization.
//! - `batch_add_authorization` - Adds an array of authorizations.
//! - `remove_authorization` - Removes an authorization.
//! - `batch_remove_authorization` - Removes an array of authorizations.
//! - `accept_authorization` - Accepts an authorization.
//! - `batch_accept_authorization` - Accepts an array of authorizations.
//! - `batch_add_signing_key_with_authorization` - Adds signing keys to target identity `id`.
//! - `revoke_offchain_authorization` - Revokes the `auth` off-chain authorization of `signer`.

#![cfg_attr(not(feature = "std"), no_std)]
#![recursion_limit = "256"]

pub mod types;
pub use types::{DidRecords as RpcDidRecords, DidStatus};

#[cfg(feature = "runtime-benchmarks")]
pub mod benchmarking;

use codec::{Decode, Encode};
use core::{
    convert::{From, TryInto},
    result::Result as StdResult,
};
use polymesh_common_utilities::{
    constants::did::{SECURITY_TOKEN, USER},
    protocol_fee::{ChargeProtocolFee, ProtocolOp},
    traits::{
        asset::AcceptTransfer,
        group::{GroupTrait, InactiveMember},
        identity::{
            AuthorizationNonce, IdentityTrait, LinkedKeyInfo, RawEvent, SigningKeyWithAuth,
            TargetIdAuthorization, Trait,
        },
        multisig::MultiSigSubTrait,
        transaction_payment::{CddAndFeeDetails, ChargeTxFee},
    },
    Context, SystematicIssuers,
};
use polymesh_primitives::{
    AuthIdentifier, Authorization, AuthorizationData, AuthorizationError, AuthorizationType, Claim,
    ClaimType, Identity as DidRecord, IdentityClaim, IdentityId, Permission, Scope, Signatory,
    SigningKey, Ticker,
};
use sp_core::sr25519::Signature;
use sp_io::hashing::blake2_256;
use sp_runtime::{
    traits::{
        AccountIdConversion, CheckedAdd, Dispatchable, Hash, IdentifyAccount, SaturatedConversion,
        Verify, Zero,
    },
    AnySignature,
};
use sp_std::{convert::TryFrom, mem::swap, prelude::*, vec};

use frame_support::{
    debug, decl_error, decl_module, decl_storage,
    dispatch::{DispatchError, DispatchResult, DispatchResultWithPostInfo},
    ensure,
    traits::{ChangeMembers, Currency, InitializeMembers},
    weights::{DispatchClass, GetDispatchInfo, Pays, Weight},
    StorageDoubleMap,
};
use frame_system::{self as system, ensure_root, ensure_signed};

pub type Event<T> = polymesh_common_utilities::traits::identity::Event<T>;

#[derive(Encode, Decode, Clone, PartialEq, Eq, Debug, PartialOrd, Ord)]
pub struct Claim1stKey {
    pub target: IdentityId,
    pub claim_type: ClaimType,
}

#[derive(Encode, Decode, Clone, PartialEq, Eq, Debug, PartialOrd, Ord)]
pub struct Claim2ndKey {
    pub issuer: IdentityId,
    pub scope: Option<Scope>,
}

#[derive(Encode, Decode, Clone, PartialEq, Eq, Debug, PartialOrd, Ord, Default)]
pub struct BatchAddClaimItem<M> {
    pub target: IdentityId,
    pub claim: Claim,
    pub expiry: Option<M>,
}

#[derive(Encode, Decode, Clone, PartialEq, Eq, Debug, PartialOrd, Ord, Default)]
pub struct BatchRevokeClaimItem {
    pub target: IdentityId,
    pub claim: Claim,
}

decl_storage! {
    trait Store for Module<T: Trait> as identity {

        /// DID -> identity info
        pub DidRecords get(fn did_records) config(): map hasher(twox_64_concat) IdentityId => DidRecord<T::AccountId>;

        /// DID -> bool that indicates if signing keys are frozen.
        pub IsDidFrozen get(fn is_did_frozen): map hasher(twox_64_concat) IdentityId => bool;

        /// It stores the current identity for current transaction.
        pub CurrentDid: Option<IdentityId>;

        /// It stores the current gas fee payer for the current transaction
        pub CurrentPayer: Option<T::AccountId>;

        /// (Target ID, claim type) (issuer,scope) -> Associated claims
        pub Claims: double_map hasher(blake2_128_concat) Claim1stKey, hasher(blake2_128_concat) Claim2ndKey => IdentityClaim;

        // Account => DID
        pub KeyToIdentityIds get(fn key_to_identity_ids) config(): map hasher(blake2_128_concat) T::AccountId => Option<LinkedKeyInfo>;

        /// Nonce to ensure unique actions. starts from 1.
        pub MultiPurposeNonce get(fn multi_purpose_nonce) build(|_| 1u64): u64;

        /// Authorization nonce per Identity. Initially is 0.
        pub OffChainAuthorizationNonce get(fn offchain_authorization_nonce): map hasher(twox_64_concat) IdentityId => AuthorizationNonce;

        /// Inmediate revoke of any off-chain authorization.
        pub RevokeOffChainAuthorization get(fn is_offchain_authorization_revoked):
            map hasher(blake2_128_concat) (Signatory<T::AccountId>, TargetIdAuthorization<T::Moment>) => bool;

        /// All authorizations that an identity/key has
        pub Authorizations: double_map hasher(blake2_128_concat)
            Signatory<T::AccountId>, hasher(twox_64_concat) u64 => Authorization<T::AccountId, T::Moment>;

        /// All authorizations that an identity has given. (Authorizer, auth_id -> authorized)
        pub AuthorizationsGiven: double_map hasher(blake2_128_concat)
            IdentityId, hasher(twox_64_concat) u64 => Signatory<T::AccountId>;

        /// It defines if authorization from a CDD provider is needed to change master key of an identity
        pub CddAuthForMasterKeyRotation get(fn cdd_auth_for_master_key_rotation): bool;
    }
    add_extra_genesis {
        config(identities): Vec<(T::AccountId, IdentityId, IdentityId, Option<u64>)>;
        config(signing_keys): Vec<(T::AccountId, IdentityId)>;
        build(|config: &GenesisConfig<T>| {
            use polymesh_common_utilities::SYSTEMATIC_ISSUERS;

            SYSTEMATIC_ISSUERS.iter()
                .for_each(|s| <Module<T>>::register_systematic_id(*s));

            // Add CDD claims to Treasury & BRR
            let sys_issuers_with_cdd = [SystematicIssuers::Treasury, SystematicIssuers::BlockRewardReserve, SystematicIssuers::Settlement];
            let id_with_cdd = sys_issuers_with_cdd.iter()
                .inspect(|iss| debug::info!( "Add Systematic CDD Claims to {}", iss))
                .map(|iss| iss.as_id())
                .collect::<Vec<_>>();

            <Module<T>>::unsafe_add_systematic_cdd_claims( &id_with_cdd, SystematicIssuers::CDDProvider);

            //  Other
            for &(ref master_account_id, issuer, did, expiry) in &config.identities {
                // Direct storage change for registering the DID and providing the claim
                assert!(!<DidRecords<T>>::contains_key(did), "Identity already exist");
                <MultiPurposeNonce>::mutate(|n| *n += 1_u64);
                let expiry = expiry.iter().map(|m| T::Moment::from(*m as u32)).next();
                <Module<T>>::unsafe_register_id(master_account_id.clone(), did);
                <Module<T>>::unsafe_add_claim( did, Claim::CustomerDueDiligence, issuer, expiry);
            }

            for &(ref signer_id, did) in &config.signing_keys {
                // Direct storage change for attaching some signing keys to identities
                assert!(<DidRecords<T>>::contains_key(did), "Identity does not exist");
                assert!(
                    <Module<T>>::can_key_be_linked_to_did(&signer_id),
                    "Signing key already linked"
                );
                <MultiPurposeNonce>::mutate(|n| *n += 1_u64);
                <Module<T>>::link_key_to_did(&signer_id, did);
                <DidRecords<T>>::mutate(did, |record| {
                    (*record).add_signing_keys(&[SigningKey::from_account_id(signer_id.clone())]);
                });
                <Module<T>>::deposit_event(RawEvent::SigningKeysAdded(
                    did,
                    [SigningKey::from_account_id(signer_id.clone())].to_vec(),
                ));
            }
        });
    }
}

decl_module! {
    /// The module declaration.
    pub struct Module<T: Trait> for enum Call where origin: T::Origin {

        type Error = Error<T>;

        // Initializing events
        // this is needed only if you are using events in your module
        fn deposit_event() = default;

        // TODO: Remove this function before mainnet. cdd_register_did should be used instead.
        /// Register a new did with a CDD claim for the caller.
        #[weight = 5_000_000_000]
        pub fn register_did(origin, signing_keys: Vec<SigningKey<T::AccountId>>) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            Self::_register_did(sender, signing_keys, Some(ProtocolOp::IdentityRegisterDid))?;
            Ok(())
        }

        /// Register `target_account` with a new Identity.
        ///
        /// # Failure
        /// - `origin` has to be a active CDD provider. Inactive CDD providers cannot add new
        /// claims.
        /// - `target_account` (master key of the new Identity) can be linked to just one and only
        /// one identity.
        /// - External signing keys can be linked to just one identity.
        ///
        /// # Weight
        /// `7_000_000_000 + 600_000 * signing_keys.len()`
        #[weight = (
            (7_000_000_000 + 600_000 * u64::try_from(signing_keys.len()).unwrap_or_default()),
            DispatchClass::Normal,
            Pays::Yes
        )]
        pub fn cdd_register_did(
            origin,
            target_account: T::AccountId,
            cdd_claim_expiry: Option<T::Moment>,
            signing_keys: Vec<SigningKey<T::AccountId>>
        ) -> DispatchResult {
            // Sender has to be part of CDDProviders
            let cdd_sender = ensure_signed(origin)?;
            let cdd_id = Context::current_identity_or::<Self>(&cdd_sender)?;

            let cdd_providers = T::CddServiceProviders::get_members();
            ensure!(cdd_providers.contains(&cdd_id), Error::<T>::UnAuthorizedCddProvider);
            // Register Identity and add claim.
            let new_id = Self::_register_did(
                target_account,
                signing_keys,
                Some(ProtocolOp::IdentityCddRegisterDid)
            )?;
            Self::unsafe_add_claim(new_id, Claim::CustomerDueDiligence, cdd_id, cdd_claim_expiry);
            Ok(())
        }

        /// It invalidates any claim generated by `cdd` from `disable_from` timestamps.
        /// You can also define an expiration time, which will invalidate all claims generated by
        /// that `cdd` and remove it as CDD member group.
        #[weight = 850_000_000]
        pub fn invalidate_cdd_claims(
            origin,
            cdd: IdentityId,
            disable_from: T::Moment,
            expiry: Option<T::Moment>,
        ) -> DispatchResult {
            ensure_root(origin)?;

            let now = <pallet_timestamp::Module<T>>::get();
            ensure!(
                T::CddServiceProviders::get_valid_members_at(now).contains(&cdd),
                Error::<T>::UnAuthorizedCddProvider);

            T::CddServiceProviders::disable_member(cdd, expiry, Some(disable_from))?;
            Self::deposit_event(RawEvent::CddClaimsInvalidated(cdd, disable_from));
            Ok(())
        }

        /// Removes specified signing keys of a DID if present.
        ///
        /// # Failure
        /// It can only called by master key owner.
        ///
        /// # Weight
        /// `950_000_000 + 60_000 * signers_to_remove.len()`
        #[weight = (
            950_000_000 + 60_000 * u64::try_from(signers_to_remove.len()).unwrap_or_default(),
            DispatchClass::Normal,
            Pays::Yes
        )]
        pub fn remove_signing_keys(origin, signers_to_remove: Vec<Signatory<T::AccountId>>) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let did = Context::current_identity_or::<Self>(&sender)?;
            let _grants_checked = Self::grant_check_only_master_key(&sender, did)?;

            // Remove links and get all authorization IDs per signer.
            let signer_and_auth_id_list = signers_to_remove.iter().filter_map(|signer| {
                if let Signatory::Account(key) = &signer {
                    if T::MultiSig::is_multisig(key) {
                        if !T::Balances::total_balance(key).is_zero() {
                            return None;
                        }
                        // Unlink multisig signers from the identity.
                        Self::unlink_multisig_signers_from_did(
                            T::MultiSig::get_key_signers(key),
                            did
                        );
                    }
                    Self::unlink_key_from_did(key, did)
                }

                // It returns the list of `auth_id` from `did`.
                let auth_ids = <Authorizations<T>>::iter_prefix_values(signer)
                    .filter_map( |authorization| {
                        if authorization.authorized_by == did {
                            Some(authorization.auth_id)
                        } else {
                            None
                        }
                    })
                    .collect::<Vec<_>>();

                Some((signer, auth_ids))
            })
            .collect::<Vec<_>>();

            // Remove authorizations
            signer_and_auth_id_list.into_iter().for_each( |(signer, auth_ids)| {
                auth_ids.into_iter().for_each( |auth_id|
                        Self::unsafe_remove_auth( signer, auth_id, &did, true));
            });

            // Update signing keys at Identity.
            <DidRecords<T>>::mutate(did, |record| {
                (*record).remove_signing_keys(&signers_to_remove);
            });

            Self::deposit_event(RawEvent::SigningKeysRemoved(did, signers_to_remove));
            Ok(())
        }

        /// Sets a new master key for a DID.
        ///
        /// # Failure
        /// Only called by master key owner.
        #[weight = 800_000_000]
        fn set_master_key(origin, new_key: T::AccountId) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let did = Context::current_identity_or::<Self>(&sender)?;
            let _grants_checked = Self::grant_check_only_master_key(&sender, did)?;

            ensure!(
                Self::can_key_be_linked_to_did(&new_key),
                Error::<T>::AlreadyLinked
            );
            T::ProtocolFee::charge_fee(ProtocolOp::IdentitySetMasterKey)?;
            <DidRecords<T>>::mutate(did,
            |record| {
                (*record).master_key = new_key.clone();
            });

            Self::deposit_event(RawEvent::MasterKeyUpdated(did, sender, new_key));
            Ok(())
        }

        /// Call this with the new master key. By invoking this method, caller accepts authorization
        /// with the new master key. If a CDD service provider approved this change, master key of
        /// the DID is updated.
        ///
        /// # Arguments
        /// * `owner_auth_id` Authorization from the owner who initiated the change
        /// * `cdd_auth_id` Authorization from a CDD service provider
        #[weight = 900_000_000]
        pub fn accept_master_key(origin, rotation_auth_id: u64, optional_cdd_auth_id: Option<u64>) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            Self::accept_master_key_rotation(sender, rotation_auth_id, optional_cdd_auth_id)
        }

        /// Set if CDD authorization is required for updating master key of an identity.
        /// Callable via root (governance)
        ///
        /// # Arguments
        /// * `auth_required` CDD Authorization required or not
        #[weight = 800_000_000]
        pub fn change_cdd_requirement_for_mk_rotation(
            origin,
            auth_required: bool,
        ) -> DispatchResult {
            ensure_root(origin)?;
            <CddAuthForMasterKeyRotation>::put(auth_required);
            Self::deposit_event(RawEvent::CddRequirementForMasterKeyUpdated(auth_required));
            Ok(())
        }

        /// Join an identity as a signing key.
        #[weight = 800_000_000]
        pub fn join_identity_as_key(origin, auth_id: u64) -> DispatchResult {
            let signer = Signatory::Account(ensure_signed(origin)?);
            Self::join_identity(signer, auth_id)
        }

        /// Join an identity as a signing identity.
        #[weight = 800_000_000]
        pub fn join_identity_as_identity(origin, auth_id: u64) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let sender_did = Context::current_identity_or::<Self>(&sender)?;
            Self::join_identity(Signatory::from(sender_did), auth_id)
        }

        /// Leave the signing key's identity.
        #[weight = 800_000_000]
        pub fn leave_identity_as_key(origin) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            if let Some(did) = Self::get_identity(&sender) {
                return Self::leave_identity(Signatory::Account(sender), did);
            }
            Ok(())
        }

        /// Leave an identity as a signing identity.
        #[weight = 800_000_000]
        pub fn leave_identity_as_identity(origin, did: IdentityId) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let sender_did = Context::current_identity_or::<Self>(&sender)?;
            Self::leave_identity(Signatory::from(sender_did), did)
        }

        /// Adds a new claim record or edits an existing one. Only called by did_issuer's signing key.
        #[weight = 950_000_000]
        pub fn add_claim(
            origin,
            target: IdentityId,
            claim: Claim,
            expiry: Option<T::Moment>,
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let issuer = Context::current_identity_or::<Self>(&sender)?;
            ensure!(<DidRecords<T>>::contains_key(target), Error::<T>::DidMustAlreadyExist);

            match claim {
                Claim::CustomerDueDiligence => Self::unsafe_add_cdd_claim(target, claim, issuer, expiry)?,
                _ => {
                    T::ProtocolFee::charge_fee(ProtocolOp::IdentityAddClaim)?;
                    Self::unsafe_add_claim(target, claim, issuer, expiry)
                }
            };
            Ok(())
        }

        /// Adds a new batch of claim records or edits an existing one. Only called by
        /// `did_issuer`'s signing key.
        ///
        /// # Weight
        /// `950_000_000 + 1_000_000 * claims.len()`
        #[weight =(
            950_000_000 + 1_000_000 * u64::try_from(claims.len()).unwrap_or_default(),
            DispatchClass::Normal,
            Pays::Yes
        )]
        pub fn batch_add_claim(
            origin,
            claims: Vec<BatchAddClaimItem<T::Moment>>
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let issuer = Context::current_identity_or::<Self>(&sender)?;
            // Check input claims.
            ensure!( claims.iter().all(
                |batch_claim_item| <DidRecords<T>>::contains_key(batch_claim_item.target)),
                Error::<T>::DidMustAlreadyExist);

            let cdd_count: usize = claims
                .iter()
                .filter(|batch_claim_item| match batch_claim_item.claim {
                    Claim::CustomerDueDiligence => true,
                    _ => false,
                })
                .count();
            if cdd_count > 0 {
                let cdd_providers = T::CddServiceProviders::get_members();
                ensure!(cdd_providers.contains(&issuer), Error::<T>::UnAuthorizedCddProvider);
            }

            T::ProtocolFee::batch_charge_fee(ProtocolOp::IdentityAddClaim, claims.len() - cdd_count)?;
            claims
                .into_iter()
                .for_each(|bci| {
                    Self::unsafe_add_claim(bci.target, bci.claim, issuer, bci.expiry)
                });
            Ok(())
        }

        /// Creates a call on behalf of another DID.
        #[weight =(
            500_000_000 + proposal.get_dispatch_info().weight,
            proposal.get_dispatch_info().class,
            Pays::Yes
        )]
        fn forwarded_call(origin, target_did: IdentityId, proposal: Box<T::Proposal>) -> DispatchResultWithPostInfo {
            let sender = ensure_signed(origin)?;

            // 1. Constraints.
            // 1.1. A valid current identity.
            let current_did = Context::current_identity::<Self>()
                .ok_or_else(||Error::<T>::MissingCurrentIdentity)?;

            // 1.2. Check that current_did is a signing key of target_did
            ensure!(
                Self::is_signer_authorized(current_did, &Signatory::Identity(target_did)),
                Error::<T>::CurrentIdentityCannotBeForwarded
            );

            // 1.3. Check that target_did has a CDD.
            ensure!(Self::has_valid_cdd(target_did), Error::<T>::TargetHasNoCdd);

            // 1.4 charge fee
            let _ = T::ChargeTxFeeTarget::charge_fee(
                proposal.encode().len().try_into().unwrap_or_default(),
                proposal.get_dispatch_info())
                    .map_err(|_| Error::<T>::FailedToChargeFee)?;

            // 2. Actions
            T::CddHandler::set_current_identity(&target_did);

            // Also set current_did roles when acting as a signing key for target_did
            // Re-dispatch call - e.g. to asset::doSomething...
            let new_origin = frame_system::RawOrigin::Signed(sender).into();

            let actual_weight = match proposal.dispatch(new_origin) {
                Ok(post_info) => post_info.actual_weight,
                Err(err) => err.post_info.actual_weight,
            };

            // If actual_weight retrieve from the proposal is `None` then refunds = 0
            // otherwise refunds = ((500_000_000 + proposal.get_dispatch_info().weight) - `actual_weight of proposal + 500_000_000`).
            Ok((actual_weight.map(|w| w + 500_000_000)).into())
        }

        /// Marks the specified claim as revoked.
        #[weight = 500_000_000]
        pub fn revoke_claim(origin,
            target: IdentityId,
            claim: Claim,
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let issuer = Context::current_identity_or::<Self>(&sender)?;
            let claim_type = claim.claim_type();
            let scope = claim.as_scope().cloned();

            Self::unsafe_revoke_claim(target, claim_type, issuer, scope);
            Ok(())
        }

        /// Revoke multiple claims in a batch.
        ///
        /// # Arguments
        /// * origin - did issuer
        /// * did_and_claim_data - Vector of the identities & the corresponding claim data whom claim needs to be revoked
        ///
        /// # Weight
        /// `500_000_000 + 150_000 * claims.len()`
        #[weight = (
            500_000_000 + 150_000 * u64::try_from(claims.len()).unwrap_or_default(),
            DispatchClass::Normal,
            Pays::Yes
        )]
        pub fn batch_revoke_claim(
            origin,
            claims: Vec<BatchRevokeClaimItem>
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let issuer = Context::current_identity_or::<Self>(&sender)?;

            claims.into_iter()
                .for_each( |bci| {
                    let claim_type = bci.claim.claim_type();
                    let scope = bci.claim.as_scope().cloned();
                    Self::unsafe_revoke_claim(bci.target, claim_type, issuer, scope)
                });
            Ok(())
        }

        /// It sets permissions for an specific `target_key` key.
        /// Only the master key of an identity is able to set signing key permissions.
        ///
        /// # Weight
        /// `600_000_000 + 300_000 * permissions.len()`
        #[weight =(
            600_000_000 + 300_000 * u64::try_from(permissions.len()).unwrap_or_default(),
            DispatchClass::Normal,
            Pays::Yes
        )]
        pub fn set_permission_to_signer(
            origin,
            signer: Signatory<T::AccountId>,
            permissions: Vec<Permission>
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let did = Context::current_identity_or::<Self>(&sender)?;
            let record = Self::grant_check_only_master_key(&sender, did)?;

            // You are trying to add a permission to did's master key. It is not needed.
            match signer {
                Signatory::Account(ref key) if record.master_key == *key => Ok(()),
                _ if record.signing_keys.iter().any(|si| si.signer == signer) => Self::update_signing_key_permissions(did, &signer, permissions),
                _ => Err(Error::<T>::InvalidSender.into()),
            }
        }

        /// It disables all signing keys at `did` identity.
        ///
        /// # Errors
        ///
        #[weight = 700_000_000]
        pub fn freeze_signing_keys(origin) -> DispatchResult {
            Self::set_frozen_signing_key_flags(origin, true)
        }

        /// Re-enables all signing keys of the caller's identity.
        #[weight = 700_000_000]
        pub fn unfreeze_signing_keys(origin) -> DispatchResult {
            Self::set_frozen_signing_key_flags(origin, false)
        }

        // TODO: Remove before mainnet launch.
        /// Emits an event with caller's identity.
        #[weight = 800_000_000]
        pub fn get_my_did(origin) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let did = Context::current_identity_or::<Self>(&sender)?;

            Self::deposit_event(RawEvent::DidStatus(did, sender));
            Ok(())
        }

        // TODO: Remove before mainnet launch.
        /// Emits an event with caller's identity and CDD status.
        #[weight = 800_000_000]
        pub fn get_cdd_of(_origin, of: T::AccountId) -> DispatchResult {
            let did_opt = Self::get_identity(&of);
            let has_cdd = did_opt.iter()
                .map(|did| Self::has_valid_cdd(*did))
                .next()
                .unwrap_or_default();

            Self::deposit_event(RawEvent::CddStatus(did_opt, of, has_cdd));
            Ok(())
        }

        // Manage generic authorizations
        /// Adds an authorization.
        #[weight = 900_000_000]
        pub fn add_authorization(
            origin,
            target: Signatory<T::AccountId>,
            authorization_data: AuthorizationData<T::AccountId>,
            expiry: Option<T::Moment>
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let from_did = Context::current_identity_or::<Self>(&sender)?;
            Self::add_auth(from_did, target, authorization_data, expiry);
            Ok(())
        }

        // Manage generic authorizations
        /// Adds an array of authorizations.
        ///
        /// # Weight
        /// `900_000_000 + 50_000 * auths.len()`
        #[weight =(
            900_000_000 + 500_000 * u64::try_from(auths.len()).unwrap_or_default(),
            DispatchClass::Normal,
            Pays::Yes
        )]
        pub fn batch_add_authorization(
            origin,
            // Vec<(target_did, auth_data, expiry)>
            auths: Vec<(Signatory<T::AccountId>, AuthorizationData<T::AccountId>, Option<T::Moment>)>
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let from_did = Context::current_identity_or::<Self>(&sender)?;
            for auth in auths {
                Self::add_auth(from_did, auth.0, auth.1, auth.2);
            }
            Ok(())
        }

        /// Removes an authorization.
        #[weight = 900_000_000]
        pub fn remove_authorization(
            origin,
            target: Signatory<T::AccountId>,
            auth_id: u64
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let from_did = Context::current_identity_or::<Self>(&sender)?;

            ensure!(
                <Authorizations<T>>::contains_key(&target, auth_id),
                Error::<T>::AuthorizationDoesNotExist
            );
            let auth = <Authorizations<T>>::get(&target, auth_id);
            let revoked = auth.authorized_by == from_did;
            ensure!(
                revoked || target.eq_either(&from_did, &sender),
                Error::<T>::Unauthorized
            );
            Self::unsafe_remove_auth(&target, auth_id, &auth.authorized_by, revoked);

            Ok(())
        }

        /// Removes an array of authorizations.
        ///
        /// # Weight
        /// `900_000_000 + 1_000_000 * auths.len()`
        #[weight =(
            900_000_000 + 1_000_000 * u64::try_from(auth_identifiers.len()).unwrap_or_default(),
            DispatchClass::Normal,
            Pays::Yes
        )]
        pub fn batch_remove_authorization(
            origin,
            auth_identifiers: Vec<AuthIdentifier<T::AccountId>>
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let from_did = Context::current_identity_or::<Self>(&sender)?;
            let mut auths = Vec::with_capacity(auth_identifiers.len());
            let mut revoked = Vec::with_capacity(auth_identifiers.len());
            for i in 0..auth_identifiers.len() {
                let auth_identifier = &auth_identifiers[i];
                ensure!(
                    <Authorizations<T>>::contains_key(&auth_identifier.0, &auth_identifier.1),
                    Error::<T>::AuthorizationDoesNotExist
                );
                auths.push(<Authorizations<T>>::get(&auth_identifier.0, &auth_identifier.1));
                revoked.push(auths[i].authorized_by == from_did);
                ensure!(
                    revoked[i] || auth_identifier.0.eq_either(&from_did, &sender),
                    Error::<T>::Unauthorized
                );
            }

            for i in 0..auth_identifiers.len() {
                let auth_identifier = &auth_identifiers[i];
                Self::unsafe_remove_auth(&auth_identifier.0, auth_identifier.1, &auths[i].authorized_by, revoked[i]);

            }

            Ok(())
        }

        /// Accepts an authorization.
        #[weight = 2_000_000_000]
        pub fn accept_authorization(
            origin,
            auth_id: u64
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let signer_key = Signatory::Account(sender.clone());
            let signer_did = Context::current_identity_or::<Self>(&sender)
                .map_or_else(
                    |_error| signer_key.clone(),
                    Signatory::from);

            // Get auth by key or by id.
            let auth_signer = if <Authorizations<T>>::contains_key(&signer_did, auth_id) {
                let auth = <Authorizations<T>>::get(&signer_did, auth_id);
                Some((auth, signer_did))
            } else if <Authorizations<T>>::contains_key(&signer_key, auth_id) {
                let auth = <Authorizations<T>>::get(&signer_key, auth_id);
                Some((auth, signer_key))
            } else {
                None
            };

            let (auth, signer) = auth_signer
                    .ok_or_else(||Error::<T>::AuthorizationDoesNotExist)?;

            match signer {
                Signatory::Identity(did) => {
                    match auth.authorization_data {
                        AuthorizationData::TransferTicker(_) =>
                            T::AcceptTransferTarget::accept_ticker_transfer(did, auth_id),
                        AuthorizationData::TransferAssetOwnership(_) =>
                            T::AcceptTransferTarget::accept_asset_ownership_transfer(did, auth_id),
                        AuthorizationData::AddMultiSigSigner(_) =>
                            T::MultiSig::accept_multisig_signer(Signatory::from(did), auth_id),
                        AuthorizationData::JoinIdentity(_) =>
                            Self::join_identity(Signatory::from(did), auth_id),
                        _ => Err(Error::<T>::UnknownAuthorization.into())
                    }
                },
                Signatory::Account(key) => {
                    match auth.authorization_data {
                        AuthorizationData::AddMultiSigSigner(_) =>
                            T::MultiSig::accept_multisig_signer(Signatory::Account(key), auth_id),
                        AuthorizationData::RotateMasterKey(_identityid) =>
                            Self::accept_master_key_rotation(key , auth_id, None),
                        AuthorizationData::JoinIdentity(_) =>
                            Self::join_identity(Signatory::Account(key), auth_id),
                        _ => Err(Error::<T>::UnknownAuthorization.into())
                    }
                }
            }
        }

        /// Accepts an array of authorizations.
        /// NB: Even if an auth is invalid (due to any reason), this batch function does NOT return an error.
        /// It will just skip that particular authorization.
        ///
        /// # Weight
        /// `2_000_000_000 + 5_000_000 * auth_ids.len()`
        #[weight =(
            2_000_000_000 + 5_000_000 * u64::try_from(auth_ids.len()).unwrap_or_default(),
            DispatchClass::Normal,
            Pays::Yes
        )]
        pub fn batch_accept_authorization(
            origin,
            auth_ids: Vec<u64>
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let signer = Context::current_identity_or::<Self>(&sender)
                .map_or_else(
                    |_error| Signatory::Account(sender),
                    Signatory::from);

            match &signer {
                Signatory::Identity(did) => {
                    for auth_id in auth_ids {

                        if <Authorizations<T>>::contains_key(&signer, auth_id) {
                            let auth = <Authorizations<T>>::get(&signer, auth_id);

                            // NB: Result is not handled, invalid auths are just ignored to let the batch function continue.
                            let _result = match auth.authorization_data {
                                AuthorizationData::TransferTicker(_) =>
                                    T::AcceptTransferTarget::accept_ticker_transfer(*did, auth_id),
                                AuthorizationData::TransferAssetOwnership(_) =>
                                    T::AcceptTransferTarget::accept_asset_ownership_transfer(*did, auth_id),
                                AuthorizationData::AddMultiSigSigner(_) =>
                                    T::MultiSig::accept_multisig_signer(Signatory::from(*did), auth_id),
                                AuthorizationData::JoinIdentity(_) =>
                                    Self::join_identity(Signatory::from(*did), auth_id),
                                _ => Err(Error::<T>::UnknownAuthorization.into())
                            };
                        }
                    }
                },
                Signatory::Account(key) => {
                    for auth_id in auth_ids {

                        if <Authorizations<T>>::contains_key(&signer, auth_id) {
                            let auth = <Authorizations<T>>::get(&signer, auth_id);

                            //NB: Result is not handled, invalid auths are just ignored to let the batch function continue.
                            let _result = match auth.authorization_data {
                                AuthorizationData::AddMultiSigSigner(_) =>
                                    T::MultiSig::accept_multisig_signer(
                                        Signatory::Account(key.clone()),
                                        auth_id
                                    ),
                                AuthorizationData::RotateMasterKey(_identityid) =>
                                    Self::accept_master_key_rotation(key.clone(), auth_id, None),
                                AuthorizationData::JoinIdentity(_) =>
                                    Self::join_identity(Signatory::Account(key.clone()), auth_id),
                                _ => Err(Error::<T>::UnknownAuthorization.into())
                            };
                        }
                    }
                }
            }

            Ok(())
        }

        /// It adds signing keys to target identity `id`.
        /// Keys are directly added to identity because each of them has an authorization.
        ///
        /// Arguments:
        ///     - `origin` Master key of `id` identity.
        ///     - `id` Identity where new signing keys will be added.
        ///     - `additional_keys` New signing items (and their authorization data) to add to target
        ///     identity.
        ///
        /// Failure
        ///     - It can only called by master key owner.
        ///     - Keys should be able to linked to any identity.
        ///
        /// # Weight
        /// `2_000_000_000 + 2_000_000 * auths.len()`
        #[weight = (
            2_000_000_000 + 2_000_000 * u64::try_from(additional_keys.len()).unwrap_or_default(),
            DispatchClass::Normal,
            Pays::Yes
        )]
        pub fn batch_add_signing_key_with_authorization(
            origin,
            additional_keys: Vec<SigningKeyWithAuth<T::AccountId>>,
            expires_at: T::Moment
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;
            let id = Context::current_identity_or::<Self>(&sender)?;
            let _grants_checked = Self::grant_check_only_master_key(&sender, id)?;

            // 0. Check expiration
            let now = <pallet_timestamp::Module<T>>::get();
            ensure!(now < expires_at, Error::<T>::AuthorizationExpired);
            let authorization = TargetIdAuthorization {
                target_id: id,
                nonce: Self::offchain_authorization_nonce(id),
                expires_at
            };
            let auth_encoded = authorization.encode();

            // 1. Verify signatures.
            for si_with_auth in additional_keys.iter() {
                let si = &si_with_auth.signing_key;

                // Get account_id from signer
                let account_id_found = match si.signer.clone() {
                    Signatory::Account(key) => Some(key),
                    Signatory::Identity(id)
                        if <DidRecords<T>>::contains_key(id) =>
                    {
                        Some(<DidRecords<T>>::get(&id).master_key)
                    },
                    _ => None
                };

                if let Some(account_id) = account_id_found {
                    if let Signatory::Account(key) = &si.signer {
                        // 1.1. Constraint 1-to-1 account to DID
                        ensure!(
                            Self::can_key_be_linked_to_did(key),
                            Error::<T>::AlreadyLinked
                        );
                    }
                    // 1.2. Offchain authorization is not revoked explicitly.
                    let si_signer_authorization = &(si.signer.clone(), authorization.clone());
                    ensure!(
                        !Self::is_offchain_authorization_revoked(si_signer_authorization),
                        Error::<T>::AuthorizationHasBeenRevoked
                    );
                    // 1.3. Verify the signature.
                    let signature = AnySignature::from(Signature::from_h512(si_with_auth.auth_signature));
                    let signer: <<AnySignature as Verify>::Signer as IdentifyAccount>::AccountId =
                        Decode::decode(&mut &account_id.encode()[..]).map_err(|_| {
                            Error::<T>::CannotDecodeSignerAccountId
                        })?;
                    ensure!(
                        signature.verify(auth_encoded.as_slice(), &signer),
                        Error::<T>::InvalidAuthorizationSignature
                    );
                } else {
                    return Err(Error::<T>::InvalidAccountKey.into());
                }
            }
            // 1.999. Charge the fee.
            T::ProtocolFee::batch_charge_fee(
                ProtocolOp::IdentityAddSigningKeysWithAuthorization,
                additional_keys.len()
            )?;
            // 2.1. Link keys to identity
            let additional_keys_si = additional_keys.into_iter()
                .map( |si_with_auth| si_with_auth.signing_key)
                .collect::<Vec<_>>();

            additional_keys_si.iter().for_each( |si| {
                if let Signatory::Account(ref key) = si.signer {
                    Self::link_key_to_did(key, id);
                }
            });
            // 2.2. Update that identity information and its offchain authorization nonce.
            <DidRecords<T>>::mutate(id, |record| {
                (*record).add_signing_keys(&additional_keys_si[..]);
            });
            <OffChainAuthorizationNonce>::mutate(id, |offchain_nonce| {
                *offchain_nonce = authorization.nonce + 1;
            });

            Self::deposit_event(RawEvent::SigningKeysAdded(id, additional_keys_si));

            Ok(())
        }

        /// It revokes the `auth` off-chain authorization of `signer`. It only takes effect if
        /// the authorized transaction is not yet executed.
        #[weight = 850_000_000]
        pub fn revoke_offchain_authorization(
            origin,
            signer: Signatory<T::AccountId>,
            auth: TargetIdAuthorization<T::Moment>
        ) -> DispatchResult {
            let sender = ensure_signed(origin)?;

            match &signer {
                Signatory::Account(key) => {
                    ensure!(&sender == key, Error::<T>::KeyNotAllowed);
                }
                Signatory::Identity(id) => {
                    ensure!(Self::is_master_key(*id, &sender), Error::<T>::NotMasterKey);
                }
            }

            Self::deposit_event(
                RawEvent::OffChainAuthorizationRevoked(
                    auth.target_id,
                    signer.clone()
                )
            );
            <RevokeOffChainAuthorization<T>>::insert((signer, auth), true);
            Ok(())
        }
    }
}

decl_error! {
    pub enum Error for Module<T: Trait> {
        /// One signing key can only belong to one DID
        AlreadyLinked,
        /// Missing current identity on the transaction
        MissingCurrentIdentity,
        /// Sender is not part of did's signing keys
        InvalidSender,
        /// No did linked to the user
        NoDIDFound,
        /// Signatory is not pre authorized by the identity
        Unauthorized,
        /// Given authorization is not pre-known
        UnknownAuthorization,
        /// Account Id cannot be extracted from signer
        InvalidAccountKey,
        /// Only CDD service providers are allowed.
        UnAuthorizedCddProvider,
        /// An invalid authorization from the owner.
        InvalidAuthorizationFromOwner,
        /// An invalid authorization from the CDD provider.
        InvalidAuthorizationFromCddProvider,
        /// The authorization to change the key was not from the owner of the master key.
        KeyChangeUnauthorized,
        /// Attestation was not by a CDD service provider.
        NotCddProviderAttestation,
        /// Authorizations are not for the same DID.
        AuthorizationsNotForSameDids,
        /// The DID must already exist.
        DidMustAlreadyExist,
        /// The Claim issuer DID must already exist.
        ClaimIssuerDidMustAlreadyExist,
        /// Sender must hold a claim issuer's signing key.
        SenderMustHoldClaimIssuerKey,
        /// Current identity cannot be forwarded, it is not a signing key of target identity.
        CurrentIdentityCannotBeForwarded,
        /// The authorization does not exist.
        AuthorizationDoesNotExist,
        /// The offchain authorization has expired.
        AuthorizationExpired,
        /// The master key is already linked to an identity.
        MasterKeyAlreadyLinked,
        /// The target DID has no valid CDD.
        TargetHasNoCdd,
        /// Authorization has been explicitly revoked.
        AuthorizationHasBeenRevoked,
        /// An invalid authorization signature.
        InvalidAuthorizationSignature,
        /// This key is not allowed to execute a given operation.
        KeyNotAllowed,
        /// Only the master key is allowed to revoke an Identity Signatory off-chain authorization.
        NotMasterKey,
        /// The DID does not exist.
        DidDoesNotExist,
        /// The DID already exists.
        DidAlreadyExists,
        /// The signing keys contain the master key.
        SigningKeysContainMasterKey,
        /// Couldn't charge fee for the transaction.
        FailedToChargeFee,
        /// Signer is not a signing key of the provided identity
        NotASigner,
        /// Cannot convert a `T::AccountId` to `AnySignature::Signer::AccountId`.
        CannotDecodeSignerAccountId,
        /// Multisig can not be unlinked from an identity while it still holds POLYX
        MultiSigHasBalance
    }
}

impl<T: Trait> Module<T> {
    /// Accepts an auth to join an identity as a signer
    pub fn join_identity(signer: Signatory<T::AccountId>, auth_id: u64) -> DispatchResult {
        ensure!(
            <Authorizations<T>>::contains_key(&signer, auth_id),
            AuthorizationError::Invalid
        );

        let auth = <Authorizations<T>>::get(&signer, auth_id);

        let permissions = match auth.authorization_data {
            AuthorizationData::JoinIdentity(permissions) => Ok(permissions),
            _ => Err(AuthorizationError::Invalid),
        }?;

        //Not really needed unless we allow identities to be deleted
        ensure!(
            <DidRecords<T>>::contains_key(&auth.authorized_by),
            "Identity does not exist"
        );

        Self::consume_auth(auth.authorized_by, signer.clone(), auth_id)?;

        Self::unsafe_join_identity(auth.authorized_by, permissions, signer)
    }

    /// Joins an identity as signer
    pub fn unsafe_join_identity(
        target_did: IdentityId,
        permissions: Vec<Permission>,
        signer: Signatory<T::AccountId>,
    ) -> DispatchResult {
        T::ProtocolFee::charge_fee(ProtocolOp::IdentityAddSigningKeysWithAuthorization)?;
        if let Signatory::Account(key) = &signer {
            ensure!(
                Self::can_key_be_linked_to_did(key),
                Error::<T>::AlreadyLinked
            );
            Self::link_key_to_did(key, target_did);
        }

        // create the SigningKey
        let sg_item = SigningKey::new(signer, permissions);

        <DidRecords<T>>::mutate(target_did, |identity| {
            identity.add_signing_keys(&[sg_item.clone()]);
        });

        Self::deposit_event(RawEvent::SigningKeysAdded(target_did, [sg_item].to_vec()));

        Ok(())
    }

    /// Adds an authorization.
    pub fn add_auth(
        from: IdentityId,
        target: Signatory<T::AccountId>,
        authorization_data: AuthorizationData<T::AccountId>,
        expiry: Option<T::Moment>,
    ) -> u64 {
        let new_nonce = Self::multi_purpose_nonce() + 1u64;
        <MultiPurposeNonce>::put(&new_nonce);

        let auth = Authorization {
            authorization_data: authorization_data.clone(),
            authorized_by: from,
            expiry,
            auth_id: new_nonce,
        };

        <Authorizations<T>>::insert(target.clone(), new_nonce, auth);
        <AuthorizationsGiven<T>>::insert(from, new_nonce, target.clone());

        // This event is split in order to help the event harvesters.
        Self::deposit_event(RawEvent::AuthorizationAdded(
            from,
            target.as_identity().cloned(),
            target.as_account().cloned(),
            new_nonce,
            authorization_data,
            expiry,
        ));

        new_nonce
    }

    /// Removes any authorization. No questions asked.
    /// NB: Please do all the required checks before calling this function.
    pub fn unsafe_remove_auth(
        target: &Signatory<T::AccountId>,
        auth_id: u64,
        authorizer: &IdentityId,
        revoked: bool,
    ) {
        <Authorizations<T>>::remove(target, auth_id);
        <AuthorizationsGiven<T>>::remove(authorizer, auth_id);
        if revoked {
            Self::deposit_event(RawEvent::AuthorizationRevoked(
                target.as_identity().cloned(),
                target.as_account().cloned(),
                auth_id,
            ));
        } else {
            Self::deposit_event(RawEvent::AuthorizationRejected(
                target.as_identity().cloned(),
                target.as_account().cloned(),
                auth_id,
            ));
        }
    }

    /// Consumes an authorization.
    /// Checks if the auth has not expired and the caller is authorized to consume this auth.
    pub fn consume_auth(
        from: IdentityId,
        target: Signatory<T::AccountId>,
        auth_id: u64,
    ) -> DispatchResult {
        ensure!(
            <Authorizations<T>>::contains_key(&target, auth_id),
            AuthorizationError::Invalid
        );

        let auth = <Authorizations<T>>::get(&target, auth_id);
        ensure!(auth.authorized_by == from, AuthorizationError::Unauthorized);
        if let Some(expiry) = auth.expiry {
            let now = <pallet_timestamp::Module<T>>::get();
            ensure!(expiry > now, AuthorizationError::Expired);
        }

        <Authorizations<T>>::remove(&target, auth_id);
        <AuthorizationsGiven<T>>::remove(auth.authorized_by, auth_id);

        Self::deposit_event(RawEvent::AuthorizationConsumed(
            target.as_identity().cloned(),
            target.as_account().cloned(),
            auth_id,
        ));
        Ok(())
    }

    /// Fetches a particular authorization.
    pub fn get_authorization(
        target: Signatory<T::AccountId>,
        auth_id: u64,
    ) -> Authorization<T::AccountId, T::Moment> {
        <Authorizations<T>>::get(target, auth_id)
    }

    /// Accepts a master key rotation.
    fn accept_master_key_rotation(
        sender: T::AccountId,
        rotation_auth_id: u64,
        optional_cdd_auth_id: Option<u64>,
    ) -> DispatchResult {
        let signer = Signatory::Account(sender.clone());
        // ensure authorization is present
        ensure!(
            <Authorizations<T>>::contains_key(&signer, rotation_auth_id),
            Error::<T>::InvalidAuthorizationFromOwner
        );

        // Accept authorization from the owner
        let rotation_auth = <Authorizations<T>>::get(&signer, rotation_auth_id);

        if let AuthorizationData::RotateMasterKey(rotation_for_did) =
            rotation_auth.authorization_data
        {
            // consume owner's authorization
            Self::consume_auth(rotation_for_did, signer, rotation_auth_id)?;
            Self::unsafe_master_key_rotation(sender, rotation_for_did, optional_cdd_auth_id)
        } else {
            Err(Error::<T>::UnknownAuthorization.into())
        }
    }

    /// Processes master key rotation.
    pub fn unsafe_master_key_rotation(
        sender: T::AccountId,
        rotation_for_did: IdentityId,
        optional_cdd_auth_id: Option<u64>,
    ) -> DispatchResult {
        // Aceept authorization from CDD service provider
        if Self::cdd_auth_for_master_key_rotation() {
            let cdd_auth_id = optional_cdd_auth_id
                .ok_or_else(|| Error::<T>::InvalidAuthorizationFromCddProvider)?;

            let signer = Signatory::Account(sender.clone());
            ensure!(
                <Authorizations<T>>::contains_key(&signer, cdd_auth_id),
                Error::<T>::InvalidAuthorizationFromCddProvider
            );
            let cdd_auth = <Authorizations<T>>::get(&signer, cdd_auth_id);

            match cdd_auth.authorization_data {
                AuthorizationData::AttestMasterKeyRotation(ref attestation_for_did) => {
                    // Attestor must be a CDD service provider
                    ensure!(
                        T::CddServiceProviders::is_member(&cdd_auth.authorized_by),
                        Error::<T>::NotCddProviderAttestation
                    );

                    // Make sure authorizations are for the same DID
                    ensure!(
                        rotation_for_did == *attestation_for_did,
                        Error::<T>::AuthorizationsNotForSameDids
                    );

                    // consume CDD service provider's authorization
                    // here we pass the known authorising identity to consume_auth, rather than the expected authorising identity
                    // as we've already checked the validity above
                    Self::consume_auth(cdd_auth.authorized_by, signer, cdd_auth_id)?;
                }
                _ => return Err(Error::<T>::UnknownAuthorization.into()),
            }
        }

        // Replace master key of the owner that initiated key rotation
        let old_master_key = Self::did_records(&rotation_for_did).master_key;
        <DidRecords<T>>::mutate(&rotation_for_did, |record| {
            Self::unlink_key_from_did(&record.master_key, rotation_for_did);
            record.master_key = sender.clone();
            Self::link_key_to_did(&sender, rotation_for_did);
        });

        Self::deposit_event(RawEvent::MasterKeyUpdated(
            rotation_for_did,
            old_master_key,
            sender,
        ));
        Ok(())
    }

    /// Updates permissions of signing items.
    /// Private and not sanitized function. It is designed to be used internally by
    /// others sanitezed functions.
    fn update_signing_key_permissions(
        target_did: IdentityId,
        signer: &Signatory<T::AccountId>,
        mut permissions: Vec<Permission>,
    ) -> DispatchResult {
        // Remove duplicates.
        permissions.sort();
        permissions.dedup();

        let mut new_s_item: Option<SigningKey<T::AccountId>> = None;

        <DidRecords<T>>::mutate(target_did, |record| {
            if let Some(mut signing_key) = (*record)
                .signing_keys
                .iter()
                .find(|si| si.signer == *signer)
                .cloned()
            {
                swap(&mut signing_key.permissions, &mut permissions);
                (*record).signing_keys.retain(|si| si.signer != *signer);
                (*record).signing_keys.push(signing_key.clone());
                new_s_item = Some(signing_key);
            }
        });

        if let Some(s) = new_s_item {
            Self::deposit_event(RawEvent::SigningPermissionsUpdated(
                target_did,
                s,
                permissions,
            ));
        }
        Ok(())
    }

    /// It checks if `key` is a signing key of `did` identity.
    /// # IMPORTANT
    /// If signing keys are frozen this function always returns false.
    /// Master key cannot be frozen.
    pub fn is_signer_authorized(did: IdentityId, signer: &Signatory<T::AccountId>) -> bool {
        let record = <DidRecords<T>>::get(did);

        // Check master id or key
        match signer {
            Signatory::Account(ref signer_key) if record.master_key == *signer_key => true,
            Signatory::Identity(ref signer_id) if did == *signer_id => true,
            _ => {
                // Check signing items if DID is not frozen.
                !Self::is_did_frozen(did)
                    && record.signing_keys.iter().any(|si| si.signer == *signer)
            }
        }
    }

    /// It checks if `key` is a signing key of `did` identity.
    pub fn is_signer(did: IdentityId, signer: &Signatory<T::AccountId>) -> bool {
        let record = <DidRecords<T>>::get(did);
        record.signing_keys.iter().any(|si| si.signer == *signer)
    }

    /// Checks if signer has correct permissions.
    fn is_signer_authorized_with_permissions(
        did: IdentityId,
        signer: &Signatory<T::AccountId>,
        permissions: Vec<Permission>,
    ) -> bool {
        let record = <DidRecords<T>>::get(did);

        match signer {
            Signatory::Account(ref signer_key) if record.master_key == *signer_key => true,
            Signatory::Identity(ref signer_id) if did == *signer_id => true,
            _ => {
                if !Self::is_did_frozen(did) {
                    if let Some(signing_key) =
                        record.signing_keys.iter().find(|&si| &si.signer == signer)
                    {
                        // It retruns true if all requested permission are in this signing item.
                        return permissions.iter().all(|required_permission| {
                            signing_key.has_permission(*required_permission)
                        });
                    }
                }
                // Signatory is not part of signing items of `did`, or
                // Did is frozen.
                false
            }
        }
    }

    /// Use `did` as reference.
    pub fn is_master_key(did: IdentityId, key: &T::AccountId) -> bool {
        key == &<DidRecords<T>>::get(did).master_key
    }

    /// It returns true if `id_claim` is not expired at `moment`.
    #[inline]
    fn is_identity_claim_not_expired_at(id_claim: &IdentityClaim, moment: T::Moment) -> bool {
        if let Some(expiry) = id_claim.expiry {
            expiry > moment.saturated_into::<u64>()
        } else {
            true
        }
    }

    /// It fetches an specific `claim_type` claim type for target identity `id`, which was issued
    /// by `issuer`.
    /// It only returns non-expired claims.
    pub fn fetch_claim(
        id: IdentityId,
        claim_type: ClaimType,
        issuer: IdentityId,
        scope: Option<Scope>,
    ) -> Option<IdentityClaim> {
        let now = <pallet_timestamp::Module<T>>::get();

        Self::fetch_base_claim_with_issuer(id, claim_type, issuer, scope)
            .into_iter()
            .find(|c| Self::is_identity_claim_not_expired_at(c, now))
    }

    /// See `Self::fetch_cdd`.
    #[inline]
    pub fn has_valid_cdd(claim_for: IdentityId) -> bool {
        let trusted_cdd_providers = T::CddServiceProviders::get_members();
        // It will never happen in production but helpful during testing.
        // TODO: Remove this condition
        if trusted_cdd_providers.is_empty() {
            return true;
        }

        Self::fetch_cdd(claim_for, T::Moment::zero()).is_some()
    }

    /// It returns the CDD identity which issued the current valid CDD claim for `claim_for`
    /// identity.
    /// # Parameters
    /// * `leeway` : This leeway is added to now() before check if claim is expired.
    ///
    /// # Safety
    ///
    /// No state change is allowed in this function because this function is used within the RPC
    /// calls.
    pub fn fetch_cdd(claim_for: IdentityId, leeway: T::Moment) -> Option<IdentityId> {
        let exp_with_leeway = <pallet_timestamp::Module<T>>::get()
            .checked_add(&leeway)
            .unwrap_or_default();

        let active_cdds = T::CddServiceProviders::get_active_members();
        let inactive_not_expired_cdds = T::CddServiceProviders::get_inactive_members()
            .into_iter()
            .filter(|cdd| !T::CddServiceProviders::is_member_expired(cdd, exp_with_leeway))
            .collect::<Vec<_>>();

        Self::fetch_base_claims(claim_for, ClaimType::CustomerDueDiligence)
            .filter(|id_claim| {
                Self::is_identity_cdd_claim_valid(
                    id_claim,
                    exp_with_leeway,
                    &active_cdds,
                    &inactive_not_expired_cdds,
                )
            })
            .map(|id_claim| id_claim.claim_issuer)
            .next()
    }

    /// A CDD claims is considered valid if:
    /// * Claim is not expired at `exp_with_leeway` moment.
    /// * Its issuer is valid, that means:
    ///   * Issuer is an active CDD provider, or
    ///   * Issuer is an inactive CDD provider but claim was updated/created before that it was
    ///   deactivated.
    fn is_identity_cdd_claim_valid(
        id_claim: &IdentityClaim,
        exp_with_leeway: T::Moment,
        active_cdds: &[IdentityId],
        inactive_not_expired_cdds: &[InactiveMember<T::Moment>],
    ) -> bool {
        Self::is_identity_claim_not_expired_at(id_claim, exp_with_leeway)
            && (active_cdds.contains(&id_claim.claim_issuer)
                || inactive_not_expired_cdds
                    .iter()
                    .filter(|cdd| cdd.id == id_claim.claim_issuer)
                    .any(|cdd| {
                        id_claim.last_update_date < cdd.deactivated_at.saturated_into::<u64>()
                    }))
    }

    /// It iterates over all claims of type `claim_type` for target `id` identity.
    /// Please note that it could return expired claims.
    fn fetch_base_claims<'a>(
        target: IdentityId,
        claim_type: ClaimType,
    ) -> impl Iterator<Item = IdentityClaim> + 'a {
        let pk = Claim1stKey { target, claim_type };
        <Claims>::iter_prefix_values(pk)
    }

    /// It fetches an specific `claim_type` claim type for target identity `id`, which was issued
    /// by `issuer`.
    fn fetch_base_claim_with_issuer(
        target: IdentityId,
        claim_type: ClaimType,
        issuer: IdentityId,
        scope: Option<Scope>,
    ) -> Option<IdentityClaim> {
        let pk = Claim1stKey { target, claim_type };
        let sk = Claim2ndKey { issuer, scope };

        if <Claims>::contains_key(&pk, &sk) {
            Some(<Claims>::get(&pk, &sk))
        } else {
            None
        }
    }

    /// It checks that `sender` is the master key of `did` Identifier and that
    /// did exists.
    /// # Return
    /// A result object containing the `DidRecord` of `did`.
    pub fn grant_check_only_master_key(
        sender: &T::AccountId,
        did: IdentityId,
    ) -> sp_std::result::Result<DidRecord<T::AccountId>, Error<T>> {
        ensure!(
            <DidRecords<T>>::contains_key(did),
            Error::<T>::DidDoesNotExist
        );
        let record = <DidRecords<T>>::get(did);
        ensure!(*sender == record.master_key, Error::<T>::KeyNotAllowed);
        Ok(record)
    }

    /// It checks if `key` is the master key or signing key of any IdentityId.
    /// Please note that _frozen signing keys_ are not lined to the frozen identity temporary.
    ///
    /// # Return
    ///
    /// An Option object containing the `IdentityId` that belongs to the key.
    pub fn get_identity(key: &T::AccountId) -> Option<IdentityId> {
        if let Some(linked_key_info) = <KeyToIdentityIds<T>>::get(key) {
            let id = match linked_key_info {
                LinkedKeyInfo::Unique(id)
                    if !Self::is_did_frozen(id) || Self::is_master_key(id, key) =>
                {
                    Some(id)
                }
                _ => None,
            };

            return id;
        }
        None
    }

    /// It freezes/unfreezes the target `did` identity.
    ///
    /// # Errors
    /// Only master key can freeze/unfreeze an identity.
    fn set_frozen_signing_key_flags(origin: T::Origin, freeze: bool) -> DispatchResult {
        let sender = ensure_signed(origin)?;
        let did = Context::current_identity_or::<Self>(&sender)?;
        let _grants_checked = Self::grant_check_only_master_key(&sender, did)?;

        if freeze {
            <IsDidFrozen>::insert(&did, true);
            Self::deposit_event(RawEvent::SigningKeysFrozen(did));
        } else {
            <IsDidFrozen>::remove(&did);
            Self::deposit_event(RawEvent::SigningKeysUnfrozen(did));
        }
        Ok(())
    }

    /// It checks that any external account can only be associated with at most one.
    /// Master keys are considered as external accounts.
    pub fn can_key_be_linked_to_did(key: &T::AccountId) -> bool {
        if <KeyToIdentityIds<T>>::get(key).is_some() {
            false
        } else {
            !T::MultiSig::is_signer(key)
        }
    }

    /// It links `key` key to `did` identity as a `key_type` type.
    /// # Errors
    /// This function can be used if `can_key_be_linked_to_did` returns true. Otherwise, it will do
    /// nothing.
    fn link_key_to_did(key: &T::AccountId, did: IdentityId) {
        if <KeyToIdentityIds<T>>::get(key).is_none() {
            // `key` is not yet linked to any identity, so no constraints.
            let linked_key_info = LinkedKeyInfo::Unique(did);
            <KeyToIdentityIds<T>>::insert(key, linked_key_info);
        }
    }

    /// It unlinks the `key` key from `did`.
    /// If there is no more associated identities, its full entry is removed.
    fn unlink_key_from_did(key: &T::AccountId, did: IdentityId) {
        if let Some(linked_key_info) = <KeyToIdentityIds<T>>::get(key) {
            match linked_key_info {
                LinkedKeyInfo::Unique(did_linked) => {
                    if did_linked == did {
                        <KeyToIdentityIds<T>>::remove(key)
                    }
                }
                LinkedKeyInfo::Group(mut dids) => {
                    dids.retain(|ref_did| *ref_did != did);
                    if dids.is_empty() {
                        <KeyToIdentityIds<T>>::remove(key);
                    } else {
                        <KeyToIdentityIds<T>>::insert(key, LinkedKeyInfo::Group(dids));
                    }
                }
            }
        }
    }

    /// It registers a did for a new asset. Only called by create_asset function.
    pub fn register_asset_did(ticker: &Ticker) -> DispatchResult {
        let did = Self::get_token_did(ticker)?;
        // Making sure there's no pre-existing entry for the DID
        // This should never happen but just being defensive here
        ensure!(
            !<DidRecords<T>>::contains_key(did),
            Error::<T>::DidAlreadyExists
        );
        <DidRecords<T>>::insert(did, DidRecord::default());
        Self::deposit_event(RawEvent::AssetDidRegistered(did, *ticker));
        Ok(())
    }

    /// IMPORTANT: No state change is allowed in this function
    /// because this function is used within the RPC calls
    /// It is a helper function that can be used to get did for any asset
    pub fn get_token_did(ticker: &Ticker) -> StdResult<IdentityId, &'static str> {
        let mut buf = Vec::new();
        buf.extend_from_slice(&SECURITY_TOKEN.encode());
        buf.extend_from_slice(&ticker.encode());
        IdentityId::try_from(T::Hashing::hash(&buf[..]).as_ref())
    }

    /// Registers a did without adding a CDD claim for it.
    pub fn _register_did(
        sender: T::AccountId,
        signing_keys: Vec<SigningKey<T::AccountId>>,
        protocol_fee_data: Option<ProtocolOp>,
    ) -> Result<IdentityId, DispatchError> {
        // Adding extrensic count to did nonce for some unpredictability
        // NB: this does not guarantee randomness
        let new_nonce =
            Self::multi_purpose_nonce() + u64::from(<system::Module<T>>::extrinsic_count()) + 7u64;
        // Even if this transaction fails, nonce should be increased for added unpredictability of dids
        <MultiPurposeNonce>::put(&new_nonce);

        // 1 Check constraints.
        // 1.1. Master key is not linked to any identity.
        ensure!(
            Self::can_key_be_linked_to_did(&sender),
            Error::<T>::MasterKeyAlreadyLinked
        );
        // 1.2. Master key is not part of signing keys.
        ensure!(
            signing_keys
                .iter()
                .find(|sk| sk.signer.as_account() == Some(&sender))
                .is_none(),
            Error::<T>::SigningKeysContainMasterKey
        );

        let block_hash = <system::Module<T>>::block_hash(<system::Module<T>>::block_number());
        let did = IdentityId::from(blake2_256(&(USER, block_hash, new_nonce).encode()));

        // 1.3. Make sure there's no pre-existing entry for the DID
        // This should never happen but just being defensive here
        ensure!(
            !<DidRecords<T>>::contains_key(did),
            Error::<T>::DidAlreadyExists
        );

        // 1.4. Signing keys can be linked to the new identity.
        for s_item in &signing_keys {
            if let Signatory::Account(ref key) = s_item.signer {
                ensure!(
                    Self::can_key_be_linked_to_did(key),
                    Error::<T>::AlreadyLinked
                );
            }
        }

        // 1.5. Charge the given fee.
        if let Some(op) = protocol_fee_data {
            T::ProtocolFee::charge_fee(op)?;
        }

        // 2. Apply changes to our extrinsic.
        // 2.1. Link master key and add pre-authorized signing keys.
        Self::link_key_to_did(&sender, did);
        let _auth_ids = signing_keys
            .iter()
            .map(|s_item| {
                Self::add_auth(
                    did,
                    s_item.signer.clone(),
                    AuthorizationData::JoinIdentity(s_item.permissions.clone()),
                    None,
                )
            })
            .collect::<Vec<_>>();

        // 2.2. Create a new identity record.
        let record = DidRecord {
            master_key: sender.clone(),
            ..Default::default()
        };
        <DidRecords<T>>::insert(&did, record);

        Self::deposit_event(RawEvent::DidCreated(did, sender, signing_keys));
        Ok(did)
    }

    /// It adds a new claim without any previous security check.
    fn unsafe_add_claim(
        target: IdentityId,
        claim: Claim,
        issuer: IdentityId,
        expiry: Option<T::Moment>,
    ) {
        let claim_type = claim.claim_type();
        let scope = claim.as_scope().cloned();
        let last_update_date = <pallet_timestamp::Module<T>>::get().saturated_into::<u64>();
        let issuance_date = Self::fetch_claim(target, claim_type, issuer, scope)
            .map_or(last_update_date, |id_claim| id_claim.issuance_date);

        let expiry = expiry.into_iter().map(|m| m.saturated_into::<u64>()).next();
        let pk = Claim1stKey { target, claim_type };
        let sk = Claim2ndKey { issuer, scope };
        let id_claim = IdentityClaim {
            claim_issuer: issuer,
            issuance_date,
            last_update_date,
            expiry,
            claim,
        };

        <Claims>::insert(&pk, &sk, id_claim.clone());
        Self::deposit_event(RawEvent::ClaimAdded(target, id_claim));
    }

    /// It ensures that CDD claim issuer is a valid CDD provider before add the claim.
    ///
    /// # Errors
    /// - 'UnAuthorizedCddProvider' is returned if `issuer` is not a CDD provider.
    fn unsafe_add_cdd_claim(
        target: IdentityId,
        claim: Claim,
        issuer: IdentityId,
        expiry: Option<T::Moment>,
    ) -> DispatchResult {
        let cdd_providers = T::CddServiceProviders::get_members();
        ensure!(
            cdd_providers.contains(&issuer),
            Error::<T>::UnAuthorizedCddProvider
        );

        Self::unsafe_add_claim(target, claim, issuer, expiry);
        Ok(())
    }

    pub fn is_identity_exists(did: &IdentityId) -> bool {
        <DidRecords<T>>::contains_key(did)
    }

    /// It removes a claim from `target` which was issued by `issuer` without any security check.
    fn unsafe_revoke_claim(
        target: IdentityId,
        claim_type: ClaimType,
        issuer: IdentityId,
        scope: Option<Scope>,
    ) {
        let pk = Claim1stKey { target, claim_type };
        let sk = Claim2ndKey { scope, issuer };
        let claim = <Claims>::get(&pk, &sk);
        <Claims>::remove(&pk, &sk);
        Self::deposit_event(RawEvent::ClaimRevoked(target, claim));
    }

    /// Returns an auth id if it is present and not expired.
    pub fn get_non_expired_auth(
        target: &Signatory<T::AccountId>,
        auth_id: &u64,
    ) -> Option<Authorization<T::AccountId, T::Moment>> {
        if !<Authorizations<T>>::contains_key(target, auth_id) {
            return None;
        }
        let auth = <Authorizations<T>>::get(target, auth_id);
        if let Some(expiry) = auth.expiry {
            let now = <pallet_timestamp::Module<T>>::get();
            if expiry > now {
                return None;
            }
        }
        Some(auth)
    }

    /// Returns identity of a signatory
    pub fn get_identity_of_signatory(signer: &Signatory<T::AccountId>) -> Option<IdentityId> {
        match signer {
            Signatory::Account(key) => Self::get_identity(&key),
            Signatory::Identity(did) => Some(*did),
        }
    }

    fn leave_identity(signer: Signatory<T::AccountId>, did: IdentityId) -> DispatchResult {
        ensure!(Self::is_signer(did, &signer), Error::<T>::NotASigner);

        if let Signatory::Account(key) = &signer {
            if T::MultiSig::is_multisig(key) {
                ensure!(
                    T::Balances::total_balance(key).is_zero(),
                    Error::<T>::MultiSigHasBalance
                );
                // Unlink multisig signers from the identity.
                Self::unlink_multisig_signers_from_did(T::MultiSig::get_key_signers(key), did);
            }
            Self::unlink_key_from_did(key, did)
        }

        // Update signing keys at Identity.
        <DidRecords<T>>::mutate(did, |record| {
            record.remove_signing_keys(&[signer.clone()]);
        });

        Self::deposit_event(RawEvent::SignerLeft(did, signer));
        Ok(())
    }

    fn unlink_multisig_signers_from_did(signers: Vec<T::AccountId>, did: IdentityId) {
        for signer in signers {
            Self::unlink_key_from_did(&signer, did)
        }
    }
}

impl<T: Trait> Module<T> {
    /// RPC call to know whether the given did has valid cdd claim or not
    pub fn is_identity_has_valid_cdd(
        target: IdentityId,
        leeway: Option<T::Moment>,
    ) -> Option<IdentityId> {
        Self::fetch_cdd(target, leeway.unwrap_or_default())
    }

    /// RPC call to query the given ticker did
    pub fn get_asset_did(ticker: Ticker) -> Result<IdentityId, &'static str> {
        Self::get_token_did(&ticker)
    }

    /// Retrieve DidRecords for `did`
    pub fn get_did_records(
        did: IdentityId,
    ) -> RpcDidRecords<T::AccountId, SigningKey<T::AccountId>> {
        if <DidRecords<T>>::contains_key(did) {
            let record = <DidRecords<T>>::get(did);
            RpcDidRecords::Success {
                master_key: record.master_key,
                signing_keys: record.signing_keys,
            }
        } else {
            RpcDidRecords::IdNotFound
        }
    }

    /// Use to get the filtered authorization data for a given signatory
    /// - if auth_type is None then return authorizations data on the basis of the `allow_expired` boolean
    /// - if auth_type is Some(value) then return filtered authorizations on the value basis type in conjunction
    ///   with `allow_expired` boolean condition
    pub fn get_filtered_authorizations(
        signatory: Signatory<T::AccountId>,
        allow_expired: bool,
        auth_type: Option<AuthorizationType>,
    ) -> Vec<Authorization<T::AccountId, T::Moment>> {
        let now = <pallet_timestamp::Module<T>>::get();

        if let Some(type_of_auth) = auth_type {
            <Authorizations<T>>::iter_prefix_values(signatory)
                .filter(|auth| {
                    if !allow_expired {
                        if let Some(expiry) = auth.expiry {
                            if expiry < now {
                                return false;
                            }
                        }
                    }
                    Self::get_type(auth.authorization_data.clone(), type_of_auth.clone())
                })
                .collect::<Vec<Authorization<T::AccountId, T::Moment>>>()
        } else {
            <Authorizations<T>>::iter_prefix_values(signatory)
                .filter(|l| {
                    if !allow_expired {
                        if let Some(expiry) = l.expiry {
                            if expiry < now {
                                return false;
                            }
                        }
                    }
                    true
                })
                .collect::<Vec<Authorization<T::AccountId, T::Moment>>>()
        }
    }

    pub fn get_type(
        authorization_data: AuthorizationData<T::AccountId>,
        type_of_auth: AuthorizationType,
    ) -> bool {
        type_of_auth
            == match authorization_data {
                AuthorizationData::AttestMasterKeyRotation(..) => {
                    AuthorizationType::AttestMasterKeyRotation
                }
                AuthorizationData::RotateMasterKey(..) => AuthorizationType::RotateMasterKey,
                AuthorizationData::TransferTicker(..) => AuthorizationType::TransferTicker,
                AuthorizationData::AddMultiSigSigner(..) => AuthorizationType::AddMultiSigSigner,
                AuthorizationData::TransferAssetOwnership(..) => {
                    AuthorizationType::TransferAssetOwnership
                }
                AuthorizationData::JoinIdentity(..) => AuthorizationType::JoinIdentity,
                AuthorizationData::Custom(..) => AuthorizationType::Custom,
                AuthorizationData::NoData => AuthorizationType::NoData,
            }
    }

    pub fn get_did_status(dids: Vec<IdentityId>) -> Vec<DidStatus> {
        let mut result = Vec::with_capacity(dids.len());
        dids.into_iter().for_each(|did| {
            // is DID exist in the ecosystem
            if !<DidRecords<T>>::contains_key(did) {
                result.push(DidStatus::Unknown);
            }
            // DID exist but whether it has valid cdd or not
            else if Self::fetch_cdd(did, T::Moment::zero()).is_some() {
                result.push(DidStatus::CddVerified);
            } else {
                result.push(DidStatus::Exists);
            }
        });
        result
    }

    /// Registers the systematic issuer with its DID.
    #[allow(dead_code)]
    fn register_systematic_id(issuer: SystematicIssuers)
    where
        <T as frame_system::Trait>::AccountId: core::fmt::Display,
    {
        let acc = issuer.as_module_id().into_account();
        let id = issuer.as_id();
        debug::info!(
            "Register Systematic id {} with account {} as {}",
            issuer,
            acc,
            id
        );

        Self::unsafe_register_id(acc, id);
    }

    /// Registers `master_key` as `id` identity.
    #[allow(dead_code)]
    fn unsafe_register_id(master_key: T::AccountId, id: IdentityId) {
        <Module<T>>::link_key_to_did(&master_key, id);
        let record = DidRecord {
            master_key: master_key.clone(),
            ..Default::default()
        };
        <DidRecords<T>>::insert(&id, record);

        Self::deposit_event(RawEvent::DidCreated(id, master_key, vec![]));
    }

    /// It returns the list of flatten identities of the given identity.
    /// It runs recursively over all signing items.
    pub fn flatten_identities(id: IdentityId, max_depth: u8) -> Vec<IdentityId> {
        if <DidRecords<T>>::contains_key(id) {
            let identity = <DidRecords<T>>::get(id);

            identity
                .signing_keys
                .into_iter()
                .flat_map(|si| match si.signer {
                    Signatory::Identity(sub_id) if max_depth > 0 => {
                        Self::flatten_identities(sub_id, max_depth - 1)
                    }
                    _ => vec![],
                })
                .chain([id].iter().cloned())
                .collect::<Vec<_>>()
        } else {
            vec![]
        }
    }

    /// Get the list of flatten keys fo the given identity.
    /// It runs recursively over all signing items.
    pub fn flatten_keys(id: IdentityId, max_depth: u8) -> Vec<T::AccountId> {
        let sub_identities = Self::flatten_identities(id, max_depth);
        sub_identities
            .into_iter()
            .flat_map(|sub_id| {
                let identity = <DidRecords<T>>::get(sub_id);
                identity
                    .signing_keys
                    .iter()
                    .filter_map(|si| match si.signer.clone() {
                        Signatory::Account(key) => Some(key),
                        _ => None,
                    })
                    .chain([identity.master_key].iter().cloned())
                    .collect::<Vec<_>>()
            })
            .collect::<Vec<_>>()
    }
}

impl<T: Trait> IdentityTrait<T::AccountId> for Module<T> {
    /// Fetches identity of a key.
    fn get_identity(key: &T::AccountId) -> Option<IdentityId> {
        Self::get_identity(key)
    }

    /// Fetches the caller's identity from the context.
    fn current_identity() -> Option<IdentityId> {
        <CurrentDid>::get()
    }

    /// Sets the caller's identity in the context.
    fn set_current_identity(id: Option<IdentityId>) {
        if let Some(id) = id {
            <CurrentDid>::put(id);
        } else {
            <CurrentDid>::kill();
        }
    }

    /// Fetches the fee payer from the context.
    fn current_payer() -> Option<T::AccountId> {
        <CurrentPayer<T>>::get()
    }

    /// Sets the fee payer in the context.
    fn set_current_payer(payer: Option<T::AccountId>) {
        if let Some(payer) = payer {
            <CurrentPayer<T>>::put(payer);
        } else {
            <CurrentPayer<T>>::kill();
        }
    }

    /// Checks if the signer is authorized.
    fn is_signer_authorized(did: IdentityId, signer: &Signatory<T::AccountId>) -> bool {
        Self::is_signer_authorized(did, signer)
    }

    /// Checks if the keys is the master key of the identity.
    fn is_master_key(did: IdentityId, key: &T::AccountId) -> bool {
        Self::is_master_key(did, key)
    }

    /// Checks if the signer is authorized and has certain permissions.
    fn is_signer_authorized_with_permissions(
        did: IdentityId,
        signer: &Signatory<T::AccountId>,
        permissions: Vec<Permission>,
    ) -> bool {
        Self::is_signer_authorized_with_permissions(did, signer, permissions)
    }

    /// Adds systematic CDD claims.
    fn unsafe_add_systematic_cdd_claims(targets: &[IdentityId], issuer: SystematicIssuers) {
        targets.iter().for_each(|new_member| {
            Self::unsafe_add_claim(
                *new_member,
                Claim::CustomerDueDiligence,
                issuer.as_id(),
                None,
            )
        });
    }

    /// Removes systematic CDD claims.
    fn unsafe_revoke_systematic_cdd_claims(targets: &[IdentityId], issuer: SystematicIssuers) {
        targets.iter().for_each(|removed_member| {
            Self::unsafe_revoke_claim(
                *removed_member,
                ClaimType::CustomerDueDiligence,
                issuer.as_id(),
                None,
            )
        });
    }

    // Provides the DID status for the given DID
    fn has_valid_cdd(target_did: IdentityId) -> bool {
        Self::has_valid_cdd(target_did)
    }
}

impl<T: Trait> ChangeMembers<IdentityId> for Module<T> {
    /// Updates systematic CDDs of members of a group.
    fn change_members_sorted(
        incoming: &[IdentityId],
        outgoing: &[IdentityId],
        _new: &[IdentityId],
    ) {
        // Add/remove Systematic CDD claims for new/removed members.
        let issuer = SystematicIssuers::CDDProvider;
        Self::unsafe_add_systematic_cdd_claims(incoming, issuer);
        Self::unsafe_revoke_systematic_cdd_claims(outgoing, issuer);
    }
}

impl<T: Trait> InitializeMembers<IdentityId> for Module<T> {
    /// Initializes members of a group by adding systematic claims for them.
    fn initialize_members(members: &[IdentityId]) {
        Self::unsafe_add_systematic_cdd_claims(members, SystematicIssuers::CDDProvider);
    }
}
